@@ -22,7 +22,6 @@
 #include <locale.h>
 #include <unistd.h>
 #include <sys/types.h>
-#include <sys/wait.h>
 
 #include <glib.h>
 #include <glib/gprintf.h>
@@ -32,7 +31,6 @@
 #include <cairo/cairo-gobject.h>
 #include <vte/vte.h>
 #include "vtepcre2.h"
-#include "glib-glue.hh"
 
 #include <algorithm>
 
@@ -42,28 +40,24 @@ class Options {
 public:
         gboolean allow_window_ops{false};
         gboolean audible_bell{false};
-        gboolean backdrop{false};
-        gboolean bold_is_bright{false};
         gboolean console{false};
         gboolean debug{false};
         gboolean icon_title{false};
         gboolean keep{false};
         gboolean no_argb_visual{false};
-        gboolean no_bidi{false};
         gboolean no_bold{false};
+        gboolean no_bold_is_bright{false};
         gboolean no_builtin_dingus{false};
         gboolean no_context_menu{false};
         gboolean no_double_buffer{false};
         gboolean no_geometry_hints{false};
         gboolean no_hyperlink{false};
         gboolean no_rewrap{false};
-        gboolean no_shaping{false};
         gboolean no_shell{false};
-        gboolean no_systemd_scope{false};
         gboolean object_notifications{false};
-        gboolean require_systemd_scope{false};
         gboolean reverse{false};
         gboolean test_mode{false};
+        gboolean use_gregex{false};
         gboolean version{false};
         gboolean whole_window_transparent{false};
         bool bg_color_set{false};
@@ -333,18 +327,14 @@ public:
                           "Allow window operations (resize, move, raise/lower, (de)iconify)", nullptr },
                         { "audible-bell", 'a', 0, G_OPTION_ARG_NONE, &audible_bell,
                           "Use audible terminal bell", nullptr },
-                        { "backdrop", 0, 0,G_OPTION_ARG_NONE, &backdrop,
-                          "Dim when toplevel unfocused", nullptr },
                         { "background-color", 0, 0, G_OPTION_ARG_CALLBACK, (void*)parse_bg_color,
-                          "Set default background color", "COLOR" },
+                          "Set default background color", nullptr },
                         { "background-image", 0, 0, G_OPTION_ARG_CALLBACK, (void*)parse_background_image,
                           "Set background image from file", "FILE" },
                         { "background-extend", 0, 0, G_OPTION_ARG_CALLBACK, (void*)parse_background_extend,
                           "Set background image extend", "EXTEND" },
                         { "blink", 0, 0, G_OPTION_ARG_CALLBACK, (void*)parse_text_blink,
                           "Text blink mode (never|focused|unfocused|always)", "MODE" },
-                        { "bold-is-bright", 'B', 0, G_OPTION_ARG_NONE, &bold_is_bright,
-                          "Bold also brightens colors", nullptr },
                         { "cell-height-scale", 0, 0, G_OPTION_ARG_DOUBLE, &cell_height_scale,
                           "Add extra line spacing", "1.0..2.0" },
                         { "cell-width-scale", 0, 0, G_OPTION_ARG_DOUBLE, &cell_width_scale,
@@ -354,9 +344,9 @@ public:
                         { "cursor-blink", 0, 0, G_OPTION_ARG_CALLBACK, (void*)parse_cursor_blink,
                           "Cursor blink mode (system|on|off)", "MODE" },
                         { "cursor-background-color", 0, 0, G_OPTION_ARG_CALLBACK, (void*)parse_cursor_bg_color,
-                          "Enable a colored cursor background", "COLOR" },
+                          "Enable a colored cursor background", nullptr },
                         { "cursor-foreground-color", 0, 0, G_OPTION_ARG_CALLBACK, (void*)parse_cursor_fg_color,
-                          "Enable a colored cursor foreground", "COLOR" },
+                          "Enable a colored cursor foreground", nullptr },
                         { "cursor-shape", 0, 0, G_OPTION_ARG_CALLBACK, (void*)parse_cursor_shape,
                           "Set cursor shape (block|underline|ibeam)", nullptr },
                         { "dingu", 'D', 0, G_OPTION_ARG_STRING_ARRAY, &dingus,
@@ -364,7 +354,7 @@ public:
                         { "debug", 'd', 0,G_OPTION_ARG_NONE, &debug,
                           "Enable various debugging checks", nullptr },
                         { "encoding", 0, 0, G_OPTION_ARG_STRING, &encoding,
-                          "Specify the terminal encoding to use", "ENCODING" },
+                          "Specify the terminal encoding to use", nullptr },
                         { "env", 0, 0, G_OPTION_ARG_FILENAME_ARRAY, &environment,
                           "Add environment variable to the child\'s environment", "VAR=VALUE" },
                         { "extra-margin", 0, 0, G_OPTION_ARG_INT, &extra_margin,
@@ -372,23 +362,25 @@ public:
                         { "font", 'f', 0, G_OPTION_ARG_STRING, &font_string,
                           "Specify a font to use", nullptr },
                         { "foreground-color", 0, 0, G_OPTION_ARG_CALLBACK, (void*)parse_fg_color,
-                          "Set default foreground color", "COLOR" },
+                          "Set default foreground color", nullptr },
+                        { "gregex", 0, 0, G_OPTION_ARG_NONE, &use_gregex,
+                          "Use GRegex instead of PCRE2", nullptr },
                         { "geometry", 'g', 0, G_OPTION_ARG_STRING, &geometry,
                           "Set the size (in characters) and position", "GEOMETRY" },
                         { "highlight-background-color", 0, 0, G_OPTION_ARG_CALLBACK, (void*)parse_hl_bg_color,
-                          "Enable distinct highlight background color for selection", "COLOR" },
+                          "Enable distinct highlight background color for selection", nullptr },
                         { "highlight-foreground-color", 0, 0, G_OPTION_ARG_CALLBACK, (void*)parse_hl_fg_color,
-                          "Enable distinct highlight foreground color for selection", "COLOR" },
+                          "Enable distinct highlight foreground color for selection", nullptr },
                         { "icon-title", 'i', 0, G_OPTION_ARG_NONE, &icon_title,
                           "Enable the setting of the icon title", nullptr },
                         { "keep", 'k', 0, G_OPTION_ARG_NONE, &keep,
                           "Live on after the command exits", nullptr },
                         { "no-argb-visual", 0, 0, G_OPTION_ARG_NONE, &no_argb_visual,
                           "Don't use an ARGB visual", nullptr },
-                        { "no-bidi", 0, 0, G_OPTION_ARG_NONE, &no_bidi,
-                          "Disable BiDi", nullptr },
                         { "no-bold", 0, 0, G_OPTION_ARG_NONE, &no_bold,
                           "Disable bold", nullptr },
+                        { "no-bold-is-bright", 'B', 0, G_OPTION_ARG_NONE, &no_bold_is_bright,
+                          "Bold does not brightens colors", nullptr },
                         { "no-builtin-dingus", 0, 0, G_OPTION_ARG_NONE, &no_builtin_dingus,
                           "Highlight URLs inside the terminal", nullptr },
                         { "no-context-menu", 0, 0, G_OPTION_ARG_NONE, &no_context_menu,
@@ -401,20 +393,14 @@ public:
                           "Disable hyperlinks", nullptr },
                         { "no-rewrap", 'R', 0, G_OPTION_ARG_NONE, &no_rewrap,
                           "Disable rewrapping on resize", nullptr },
-                        { "no-shaping", 0, 0, G_OPTION_ARG_NONE, &no_shaping,
-                          "Disable Arabic shaping", nullptr },
                         { "no-shell", 'S', 0, G_OPTION_ARG_NONE, &no_shell,
                           "Disable spawning a shell inside the terminal", nullptr },
-                        { "no-systemd-scope", 0, 0, G_OPTION_ARG_NONE, &no_systemd_scope,
-                          "Don't use systemd user scope", nullptr },
                         { "object-notifications", 'N', 0, G_OPTION_ARG_NONE, &object_notifications,
                           "Print VteTerminal object notifications", nullptr },
                         { "output-file", 0, 0, G_OPTION_ARG_FILENAME, &output_filename,
                           "Save terminal contents to file at exit", nullptr },
                         { "reverse", 0, 0, G_OPTION_ARG_NONE, &reverse,
                           "Reverse foreground/background colors", nullptr },
-                        { "require-systemd-scope", 0, 0, G_OPTION_ARG_NONE, &require_systemd_scope,
-                          "Require use of a systemd user scope", nullptr },
                         { "scrollback-lines", 'n', 0, G_OPTION_ARG_INT, &scrollback_lines,
                           "Specify the number of scrollback-lines (-1 for infinite)", nullptr },
                         { "transparent", 'T', 0, G_OPTION_ARG_INT, &transparency_percent,
@@ -515,15 +501,28 @@ verbose_fprintf(FILE* fp,
 
 /* regex */
 
+static GRegex*
+compile_gregex(char const* pattern,
+               bool caseless,
+               GError** error)
+{
+        GRegexCompileFlags flags = GRegexCompileFlags(G_REGEX_OPTIMIZE |
+                                                      G_REGEX_MULTILINE |
+                                                      caseless ? G_REGEX_CASELESS : 0);
+
+        return g_regex_new(pattern, flags, GRegexMatchFlags(0), error);
+}
+
 static void
 jit_regex(VteRegex* regex,
           char const* pattern)
 {
-        auto error = vte::glib::Error{};
-        if (!vte_regex_jit(regex, PCRE2_JIT_COMPLETE, error) ||
-            !vte_regex_jit(regex, PCRE2_JIT_PARTIAL_SOFT, error)) {
-                if (!error.matches(VTE_REGEX_ERROR, -45 /* PCRE2_ERROR_JIT_BADOPTION: JIT not supported */))
-                        verbose_printerr("JITing regex \"%s\" failed: %s\n", pattern, error.message());
+        GError *err = nullptr;
+        if (!vte_regex_jit(regex, PCRE2_JIT_COMPLETE, &err) ||
+            !vte_regex_jit(regex, PCRE2_JIT_PARTIAL_SOFT, &err)) {
+                if (!g_error_matches(err, VTE_REGEX_ERROR, -45 /* PCRE2_ERROR_JIT_BADOPTION: JIT not supported */))
+                        verbose_printerr("JITing regex \"%s\" failed: %s\n", pattern, err->message);
+                g_error_free(err);
         }
 }
 
@@ -643,20 +642,29 @@ vteapp_search_popover_update_regex(VteappSearchPopover* popover)
         popover->regex_pattern = nullptr;
 
         if (search_text[0] != '\0') {
-                auto error = vte::glib::Error{};
-                auto regex = compile_regex_for_search(pattern, caseless, error);
-                vte_terminal_search_set_regex(popover->terminal, regex, 0);
-                if (regex != nullptr)
-                        vte_regex_unref(regex);
+                GError* error = nullptr;
 
-                if (error.error()) {
+                if (options.use_gregex) {
+                        auto regex = compile_gregex(pattern, caseless, &error);
+                        vte_terminal_search_set_gregex(popover->terminal, regex, GRegexMatchFlags(0));
+                        if (regex != nullptr)
+                                g_regex_unref(regex);
+                } else {
+                        auto regex = compile_regex_for_search(pattern, caseless, &error);
+                        vte_terminal_search_set_regex(popover->terminal, regex, 0);
+                        if (regex != nullptr)
+                                vte_regex_unref(regex);
+                }
+
+                if (error == nullptr) {
                         popover->has_regex = true;
                         popover->regex_pattern = pattern; /* adopt */
                         pattern = nullptr; /* adopted */
                         gtk_widget_set_tooltip_text(popover->search_entry, nullptr);
                 } else {
                         popover->has_regex = false;
-                        gtk_widget_set_tooltip_text(popover->search_entry, error.message());
+                        gtk_widget_set_tooltip_text(popover->search_entry, error->message);
+                        g_error_free(error);
                 }
         }
 
@@ -705,8 +713,10 @@ vteapp_search_popover_init(VteappSearchPopover* popover)
                                popover->revealer, "reveal-child",
                                GBindingFlags(G_BINDING_DEFAULT | G_BINDING_SYNC_CREATE));
 
+#if GTK_CHECK_VERSION(3, 16, 0)
         g_signal_connect_swapped(popover->search_entry, "next-match", G_CALLBACK(vteapp_search_popover_search_forward), popover);
         g_signal_connect_swapped(popover->search_entry, "previous-match", G_CALLBACK(vteapp_search_popover_search_backward), popover);
+#endif
         g_signal_connect_swapped(popover->search_entry, "search-changed", G_CALLBACK(vteapp_search_popover_update_regex), popover);
 
         g_signal_connect_swapped(popover->search_next_button,"clicked", G_CALLBACK(vteapp_search_popover_search_forward), popover);
@@ -804,8 +814,6 @@ struct _VteappTerminal {
         VteTerminal parent;
 
         cairo_pattern_t* background_pattern;
-        bool has_backdrop;
-        bool use_backdrop;
 };
 
 struct _VteappTerminalClass {
@@ -816,8 +824,6 @@ static GType vteapp_terminal_get_type(void);
 
 G_DEFINE_TYPE(VteappTerminal, vteapp_terminal, VTE_TYPE_TERMINAL)
 
-#define BACKDROP_ALPHA (0.2)
-
 static void
 vteapp_terminal_realize(GtkWidget* widget)
 {
@@ -860,8 +866,7 @@ vteapp_terminal_draw(GtkWidget* widget,
                 cairo_rectangle(cr, 0.0, 0.0,
                                 gtk_widget_get_allocated_width(widget),
                                 gtk_widget_get_allocated_height(widget));
-                GdkRGBA bg;
-                vte_terminal_get_color_background_for_draw(VTE_TERMINAL(terminal), &bg);
+                auto bg = options.get_color_bg();
                 cairo_set_source_rgba(cr, bg.red, bg.green, bg.blue, 1.0);
                 cairo_paint(cr);
 
@@ -871,34 +876,11 @@ vteapp_terminal_draw(GtkWidget* widget,
                 cairo_paint(cr);
 
                 cairo_pop_group_to_source(cr);
-                cairo_paint_with_alpha(cr, bg.alpha);
+                cairo_paint_with_alpha(cr, options.get_alpha());
 
         }
 
-        auto rv = GTK_WIDGET_CLASS(vteapp_terminal_parent_class)->draw(widget, cr);
-
-        if (terminal->use_backdrop && terminal->has_backdrop) {
-                cairo_set_operator(cr, CAIRO_OPERATOR_OVER);
-                cairo_set_source_rgba(cr, 0, 0, 0, BACKDROP_ALPHA);
-                cairo_rectangle(cr, 0.0, 0.0,
-                                gtk_widget_get_allocated_width(widget),
-                                gtk_widget_get_allocated_height(widget));
-                cairo_paint(cr);
-        }
-
-        return rv;
-}
-
-static void
-vteapp_terminal_style_updated(GtkWidget* widget)
-{
-        GTK_WIDGET_CLASS(vteapp_terminal_parent_class)->style_updated(widget);
-
-        auto context = gtk_widget_get_style_context(widget);
-        auto flags = gtk_style_context_get_state(context);
-
-        VteappTerminal* terminal = VTEAPP_TERMINAL(widget);
-        terminal->has_backdrop = (flags & GTK_STATE_FLAG_BACKDROP) != 0;
+        return GTK_WIDGET_CLASS(vteapp_terminal_parent_class)->draw(widget, cr);
 }
 
 static void
@@ -908,15 +890,12 @@ vteapp_terminal_class_init(VteappTerminalClass *klass)
         widget_class->realize = vteapp_terminal_realize;
         widget_class->unrealize = vteapp_terminal_unrealize;
         widget_class->draw = vteapp_terminal_draw;
-        widget_class->style_updated = vteapp_terminal_style_updated;
 }
 
 static void
 vteapp_terminal_init(VteappTerminal *terminal)
 {
         terminal->background_pattern = nullptr;
-        terminal->has_backdrop = false;
-        terminal->use_backdrop = options.backdrop;
 
         if (options.background_pixbuf != nullptr)
                 vte_terminal_set_clear_background(VTE_TERMINAL(terminal), false);
@@ -959,8 +938,6 @@ struct _VteappWindow {
         GPid child_pid;
         GtkWidget* search_popover;
 
-        bool fullscreen{false};
-
         /* used for updating the geometry hints */
         int cached_cell_width{0};
         int cached_cell_height{0};
@@ -991,17 +968,26 @@ vteapp_window_add_dingus(VteappWindow* window,
                          char const* const* dingus)
 {
         for (auto i = 0; dingus[i] != nullptr; i++) {
-                auto tag = int{-1};
-                auto error = vte::glib::Error{};
-                auto regex = compile_regex_for_match(dingus[i], true, error);
-                if (regex) {
-                        tag = vte_terminal_match_add_regex(window->terminal, regex, 0);
-                        vte_regex_unref(regex);
+                int tag = -1;
+                GError* error = nullptr;
+                if (options.use_gregex) {
+                        auto regex = compile_gregex(dingus[i], true, &error);
+                        if (regex) {
+                                tag = vte_terminal_match_add_gregex(window->terminal, regex,
+                                                                    GRegexMatchFlags(0));
+                                g_regex_unref(regex);
+                        }
+                } else {
+                        auto regex = compile_regex_for_match(dingus[i], true, &error);
+                        if (regex) {
+                                tag = vte_terminal_match_add_regex(window->terminal, regex, 0);
+                                vte_regex_unref(regex);
+                        }
                 }
 
-                if (error.error()) {
-                        verbose_printerr("Failed to compile regex \"%s\": %s\n",
-                                         dingus[i], error.message());
+                if (error != nullptr) {
+                        verbose_printerr("Failed to compile regex \"%s\": %s\n", dingus[i], error->message);
+                        g_error_free(error);
                 }
 
                 if (tag != -1)
@@ -1067,7 +1053,11 @@ vteapp_window_update_geometry(VteappWindow* window)
                         geometry.min_height = geometry.base_height + cell_height * MIN_ROWS;
 
                         gtk_window_set_geometry_hints(GTK_WINDOW(window),
+#if GTK_CHECK_VERSION (3, 19, 5)
                                                       nullptr,
+#else
+                                                      terminal_widget,
+#endif
                                                       &geometry,
                                                       GdkWindowHints(GDK_HINT_RESIZE_INC |
                                                                      GDK_HINT_MIN_SIZE |
@@ -1100,9 +1090,7 @@ vteapp_window_resize(VteappWindow* window)
         /* Don't do this for maximised or tiled windows. */
         auto win = gtk_widget_get_window(GTK_WIDGET(window));
         if (win != nullptr &&
-            (gdk_window_get_state(win) & (GDK_WINDOW_STATE_MAXIMIZED |
-                                          GDK_WINDOW_STATE_FULLSCREEN |
-                                          GDK_WINDOW_STATE_TILED)) != 0)
+            (gdk_window_get_state(win) & (GDK_WINDOW_STATE_MAXIMIZED | GDK_WINDOW_STATE_TILED)) != 0)
                 return;
 
         /* First, update the geometry hints, so that the cached_* members are up-to-date */
@@ -1207,15 +1195,12 @@ vteapp_window_launch_argv(VteappWindow* window,
                           char** argv,
                           GError** error)
 {
-        auto const spawn_flags = GSpawnFlags(G_SPAWN_SEARCH_PATH_FROM_ENVP |
-                                             (options.no_systemd_scope ? VTE_SPAWN_NO_SYSTEMD_SCOPE : 0) |
-                                             (options.require_systemd_scope ? VTE_SPAWN_REQUIRE_SYSTEMD_SCOPE : 0));
         vte_terminal_spawn_async(window->terminal,
                                  VTE_PTY_DEFAULT,
                                  options.working_directory,
                                  argv,
                                  options.environment,
-                                 spawn_flags,
+                                 G_SPAWN_SEARCH_PATH_FROM_ENVP,
                                  nullptr, nullptr, nullptr, /* child setup, data and destroy */
                                  30 * 1000 /* 30s timeout */,
                                  nullptr /* cancellable */,
@@ -1305,20 +1290,22 @@ vteapp_window_fork(VteappWindow* window,
 static void
 vteapp_window_launch(VteappWindow* window)
 {
-        auto rv = bool{};
-        auto error = vte::glib::Error{};
+        bool rv;
+        GError* err = nullptr;
 
         if (options.exec_argv != nullptr)
-                rv = vteapp_window_launch_argv(window, options.exec_argv, error);
+                rv = vteapp_window_launch_argv(window, options.exec_argv, &err);
         else if (options.command != nullptr)
-                rv = vteapp_window_launch_commandline(window, options.command, error);
+                rv = vteapp_window_launch_commandline(window, options.command, &err);
         else if (!options.no_shell)
-                rv = vteapp_window_launch_shell(window, error);
+                rv = vteapp_window_launch_shell(window, &err);
         else
-                rv = vteapp_window_fork(window, error);
+                rv = vteapp_window_fork(window, &err);
 
-        if (!rv)
-                verbose_printerr("Error launching: %s\n", error.message());
+        if (!rv) {
+                verbose_printerr("Error launching: %s\n", err->message);
+                g_error_free(err);
+        }
 }
 
 static void
@@ -1408,25 +1395,6 @@ window_action_find_cb(GSimpleAction* action,
         gtk_toggle_button_set_active(window->find_button, true);
 }
 
-
-static void
-window_action_fullscreen_state_cb (GSimpleAction *action,
-                                   GVariant *state,
-                                   void* data)
-{
-        VteappWindow* window = VTEAPP_WINDOW(data);
-
-        if (!gtk_widget_get_realized(GTK_WIDGET(window)))
-                return;
-
-        if (g_variant_get_boolean(state))
-                gtk_window_fullscreen(GTK_WINDOW(window));
-        else
-                gtk_window_unfullscreen(GTK_WINDOW(window));
-
-        /* The window-state-changed callback will update the action's actual state */
-}
-
 static bool
 vteapp_window_show_context_menu(VteappWindow* window,
                                 guint button,
@@ -1462,43 +1430,30 @@ vteapp_window_show_context_menu(VteappWindow* window,
                 }
 
                 /* Test extra match API */
-                static const char extra_pattern[] = "(\\d+)\\s*(\\w+)";
                 char* extra_match = nullptr;
-                char *extra_subst = nullptr;
-                auto error = vte::glib::Error{};
-                auto regex = compile_regex_for_match(extra_pattern, false, error);
-                error.assert_no_error();
-                vte_terminal_event_check_regex_simple(window->terminal, event,
-                                                      &regex, 1, 0,
-                                                      &extra_match);
-
-                if (extra_match != nullptr &&
-                    (extra_subst = vte_regex_substitute(regex, extra_match, "$2 $1",
-                                                        PCRE2_SUBSTITUTE_EXTENDED |
-                                                        PCRE2_SUBSTITUTE_GLOBAL,
-                                                        error)) == nullptr) {
-                        verbose_printerr("Substitution failed: %s\n", error.message());
+                if (options.use_gregex) {
+                        auto regex = compile_gregex("\\d+", false, nullptr);
+                        vte_terminal_event_check_gregex_simple(window->terminal, event,
+                                                               &regex, 1,
+                                                               GRegexMatchFlags(0),
+                                                               &extra_match);
+                        g_regex_unref(regex);
+                } else {
+                        auto regex = compile_regex_for_match("\\d+", false, nullptr);
+                        vte_terminal_event_check_regex_simple(window->terminal, event,
+                                                              &regex, 1, 0,
+                                                              &extra_match);
+                        vte_regex_unref(regex);
                 }
 
-                vte_regex_unref(regex);
-
                 if (extra_match != nullptr) {
-                        if (extra_subst != nullptr)
-                                verbose_print("%s match: %s => %s\n", extra_pattern, extra_match, extra_subst);
-                        else
-                                verbose_print("%s match: %s\n", extra_pattern, extra_match);
+                        verbose_print("\\d+ match: %s\n", extra_match);
+                        g_free(extra_match);
                 }
-                g_free(hyperlink);
-                g_free(match);
-                g_free(extra_match);
-                g_free(extra_subst);
         }
 
         g_menu_append(menu, "_Paste", "win.paste");
 
-        if (window->fullscreen)
-                g_menu_append(menu, "_Fullscreen", "win.fullscreen");
-
         auto popup = gtk_menu_new_from_model(G_MENU_MODEL(menu));
         gtk_menu_attach_to_widget(GTK_MENU(popup), GTK_WIDGET(window->terminal), nullptr);
         gtk_menu_popup(GTK_MENU(popup), nullptr, nullptr, nullptr, nullptr, button, timestamp);
@@ -1541,29 +1496,26 @@ window_child_exited_cb(VteTerminal* term,
                        int status,
                        VteappWindow* window)
 {
-        if (WIFEXITED(status))
-                verbose_printerr("Child exited with status %x\n", WEXITSTATUS(status));
-        else if (WIFSIGNALED(status))
-                verbose_printerr("Child terminated by signal %d\n", WTERMSIG(status));
-        else
-                verbose_printerr("Child terminated\n");
+        verbose_printerr("Child exited with status %x\n", status);
 
         if (options.output_filename != nullptr) {
                 auto file = g_file_new_for_commandline_arg(options.output_filename);
-                auto error = vte::glib::Error{};
-                auto stream = g_file_replace(file, nullptr, false, G_FILE_CREATE_NONE, nullptr, error);
+
+                GError* err = nullptr;
+                auto stream = g_file_replace(file, nullptr, false, G_FILE_CREATE_NONE, nullptr, &err);
                 g_object_unref(file);
                 if (stream != nullptr) {
                         vte_terminal_write_contents_sync(window->terminal,
                                                          G_OUTPUT_STREAM(stream),
                                                          VTE_WRITE_DEFAULT,
-                                                         nullptr, error);
+                                                         nullptr, &err);
                         g_object_unref(stream);
                 }
 
-                if (error.error()) {
+                if (err != nullptr) {
                         verbose_printerr("Failed to write output to \"%s\": %s\n",
-                                         options.output_filename, error.message());
+                                         options.output_filename, err->message);
+                        g_error_free(err);
                 }
         }
 
@@ -1805,7 +1757,6 @@ vteapp_window_constructed(GObject *object)
                 { "paste",       window_action_paste_cb,  nullptr, nullptr, nullptr },
                 { "reset",       window_action_reset_cb,      "b", nullptr, nullptr },
                 { "find",        window_action_find_cb,   nullptr, nullptr, nullptr },
-                { "fullscreen",  nullptr,                 nullptr, "false", window_action_fullscreen_state_cb },
         };
 
         GActionMap* map = G_ACTION_MAP(window);
@@ -1865,16 +1816,15 @@ vteapp_window_constructed(GObject *object)
                 g_signal_connect(window->terminal, "notify", G_CALLBACK(window_notify_cb), window);
 
         /* Settings */
-        if (options.no_double_buffer) {
-                G_GNUC_BEGIN_IGNORE_DEPRECATIONS;
+        if (options.no_double_buffer)
                 gtk_widget_set_double_buffered(GTK_WIDGET(window->terminal), false);
-                G_GNUC_END_IGNORE_DEPRECATIONS;
-        }
 
         if (options.encoding != nullptr) {
-                auto error = vte::glib::Error{};
-                if (!vte_terminal_set_encoding(window->terminal, options.encoding, error))
-                        g_printerr("Failed to set encoding: %s\n", error.message());
+                GError* err = nullptr;
+                if (!vte_terminal_set_encoding(window->terminal, options.encoding, &err)) {
+                        verbose_printerr("Failed to set encoding: %s\n", err->message);
+                        g_error_free(err);
+                }
         }
 
         if (options.word_char_exceptions != nullptr)
@@ -1883,14 +1833,12 @@ vteapp_window_constructed(GObject *object)
         vte_terminal_set_allow_hyperlink(window->terminal, !options.no_hyperlink);
         vte_terminal_set_audible_bell(window->terminal, options.audible_bell);
         vte_terminal_set_allow_bold(window->terminal, !options.no_bold);
-        vte_terminal_set_bold_is_bright(window->terminal, options.bold_is_bright);
+        vte_terminal_set_bold_is_bright(window->terminal, !options.no_bold_is_bright);
         vte_terminal_set_cell_height_scale(window->terminal, options.cell_height_scale);
         vte_terminal_set_cell_width_scale(window->terminal, options.cell_width_scale);
         vte_terminal_set_cjk_ambiguous_width(window->terminal, options.cjk_ambiguous_width);
         vte_terminal_set_cursor_blink_mode(window->terminal, options.cursor_blink_mode);
         vte_terminal_set_cursor_shape(window->terminal, options.cursor_shape);
-        vte_terminal_set_enable_bidi(window->terminal, !options.no_bidi);
-        vte_terminal_set_enable_shaping(window->terminal, !options.no_shaping);
         vte_terminal_set_mouse_autohide(window->terminal, true);
         vte_terminal_set_rewrap_on_resize(window->terminal, !options.no_rewrap);
         vte_terminal_set_scroll_on_output(window->terminal, false);
@@ -1993,22 +1941,6 @@ vteapp_window_style_updated(GtkWidget* widget)
         vteapp_window_resize(window);
 }
 
-static gboolean
-vteapp_window_state_event (GtkWidget* widget,
-                           GdkEventWindowState* event)
-{
-        VteappWindow* window = VTEAPP_WINDOW(widget);
-
-        if (event->changed_mask & GDK_WINDOW_STATE_FULLSCREEN) {
-                window->fullscreen = (event->new_window_state & GDK_WINDOW_STATE_FULLSCREEN) != 0;
-
-                auto action = reinterpret_cast<GSimpleAction*>(g_action_map_lookup_action(G_ACTION_MAP(window), "fullscreen"));
-                g_simple_action_set_state(action, g_variant_new_boolean (window->fullscreen));
-        }
-
-        return GTK_WIDGET_CLASS(vteapp_window_parent_class)->window_state_event(widget, event);
-}
-
 static void
 vteapp_window_class_init(VteappWindowClass* klass)
 {
@@ -2020,7 +1952,6 @@ vteapp_window_class_init(VteappWindowClass* klass)
         widget_class->realize = vteapp_window_realize;
         widget_class->show = vteapp_window_show;
         widget_class->style_updated = vteapp_window_style_updated;
-        widget_class->window_state_event = vteapp_window_state_event;
 
         gtk_widget_class_set_template_from_resource(widget_class, "/org/gnome/vte/app/ui/window.ui");
 
@@ -2150,7 +2081,9 @@ main(int argc,
         setlocale(LC_ALL, "");
 
 	bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
+#ifdef HAVE_DECL_BIND_TEXTDOMAIN_CODESET
 	bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
+#endif
 
         /* Not interested in silly debug spew, bug #749195 */
         if (g_getenv("G_ENABLE_DIAGNOSTIC") == nullptr)
@@ -2159,9 +2092,10 @@ main(int argc,
        g_set_prgname("Terminal");
        g_set_application_name("Terminal");
 
-       auto error = vte::glib::Error{};
-       if (!options.parse_argv(argc, argv, error)) {
-               g_printerr("Error parsing arguments: %s\n", error.message());
+       GError* error = nullptr;
+       if (!options.parse_argv(argc, argv, &error)) {
+               g_printerr("Error parsing arguments: %s\n", error->message);
+               g_error_free(error);
                return EXIT_FAILURE;
        }
 
@@ -2169,7 +2103,7 @@ main(int argc,
                 verbose_printerr("VTE_CJK_WIDTH is not supported anymore, use --cjk-width instead\n");
 
        if (options.version) {
-               g_print("VTE Application %s %s\n", VERSION, vte_get_features());
+               g_print("VTE Application %s\n", VERSION);
                return EXIT_SUCCESS;
        }
 
@@ -2177,8 +2111,10 @@ main(int argc,
                gdk_window_set_debug_updates(true);
 #ifdef VTE_DEBUG
        if (options.test_mode) {
-               vte_set_test_flags(VTE_TEST_FLAGS_ALL);
+               g_setenv("VTE_TEST", "1", true);
                options.allow_window_ops = true;
+       } else {
+               g_unsetenv("VTE_TEST");
        }
 #endif
 
